@@ -4,8 +4,9 @@ use 5.006_000;
 
 use strict;
 use warnings;
+no warnings 'recursion';
 
-our $VERSION = q(1.0012);
+our $VERSION = q(1.0013);
 
 # This is to allow DBM::Deep::Array to handle negative indices on
 # its own. Otherwise, Perl would intercept the call to negative
@@ -35,7 +36,7 @@ sub FETCH {
     my $self = shift->_get_self;
     my ($key) = @_;
 
-    $self->lock( $self->LOCK_SH );
+    $self->lock_shared;
 
     if ( !defined $key ) {
         $self->unlock;
@@ -66,7 +67,7 @@ sub STORE {
     my $self = shift->_get_self;
     my ($key, $value) = @_;
 
-    $self->lock( $self->LOCK_EX );
+    $self->lock_exclusive;
 
     my $size;
     my $idx_is_numeric;
@@ -107,7 +108,7 @@ sub EXISTS {
     my $self = shift->_get_self;
     my ($key) = @_;
 
-    $self->lock( $self->LOCK_SH );
+    $self->lock_shared;
 
     if ( !defined $key ) {
         $self->unlock;
@@ -139,7 +140,7 @@ sub DELETE {
     my ($key) = @_;
     warn "ARRAY::DELETE($self,$key)\n" if DBM::Deep::DEBUG;
 
-    $self->lock( $self->LOCK_EX );
+    $self->lock_exclusive;
 
     my $size = $self->FETCHSIZE;
     if ( !defined $key ) {
@@ -177,14 +178,14 @@ sub DELETE {
 sub FETCHSIZE {
     my $self = shift->_get_self;
 
-    $self->lock( $self->LOCK_SH );
+    $self->lock_shared;
 
-    my $SAVE_FILTER = $self->_storage->{filter_fetch_value};
-    $self->_storage->{filter_fetch_value} = undef;
+    my $SAVE_FILTER = $self->_engine->storage->{filter_fetch_value};
+    $self->_engine->storage->{filter_fetch_value} = undef;
 
     my $size = $self->FETCH('length') || 0;
 
-    $self->_storage->{filter_fetch_value} = $SAVE_FILTER;
+    $self->_engine->storage->{filter_fetch_value} = $SAVE_FILTER;
 
     $self->unlock;
 
@@ -195,14 +196,14 @@ sub STORESIZE {
     my $self = shift->_get_self;
     my ($new_length) = @_;
 
-    $self->lock( $self->LOCK_EX );
+    $self->lock_exclusive;
 
-    my $SAVE_FILTER = $self->_storage->{filter_store_value};
-    $self->_storage->{filter_store_value} = undef;
+    my $SAVE_FILTER = $self->_engine->storage->{filter_store_value};
+    $self->_engine->storage->{filter_store_value} = undef;
 
     my $result = $self->STORE('length', $new_length, 'length');
 
-    $self->_storage->{filter_store_value} = $SAVE_FILTER;
+    $self->_engine->storage->{filter_store_value} = $SAVE_FILTER;
 
     $self->unlock;
 
@@ -212,7 +213,7 @@ sub STORESIZE {
 sub POP {
     my $self = shift->_get_self;
 
-    $self->lock( $self->LOCK_EX );
+    $self->lock_exclusive;
 
     my $length = $self->FETCHSIZE();
 
@@ -233,7 +234,7 @@ sub POP {
 sub PUSH {
     my $self = shift->_get_self;
 
-    $self->lock( $self->LOCK_EX );
+    $self->lock_exclusive;
 
     my $length = $self->FETCHSIZE();
 
@@ -260,7 +261,7 @@ sub SHIFT {
     my $self = shift->_get_self;
     warn "SHIFT($self)\n" if DBM::Deep::DEBUG;
 
-    $self->lock( $self->LOCK_EX );
+    $self->lock_exclusive;
 
     my $length = $self->FETCHSIZE();
 
@@ -289,7 +290,7 @@ sub UNSHIFT {
     my $self = shift->_get_self;
     my @new_elements = @_;
 
-    $self->lock( $self->LOCK_EX );
+    $self->lock_exclusive;
 
     my $length = $self->FETCHSIZE();
     my $new_size = scalar @new_elements;
@@ -314,7 +315,7 @@ sub UNSHIFT {
 sub SPLICE {
     my $self = shift->_get_self;
 
-    $self->lock( $self->LOCK_EX );
+    $self->lock_exclusive;
 
     my $length = $self->FETCHSIZE();
 
